"""
Package for the types of options available to option-handlers.
"""
from ._ClassOption import ClassOption
from ._CountOption import CountOption
from ._FlagOption import FlagOption
from ._Option import Option
from ._TypedOption import TypedOption
