"""
Module for constants related to abstract base-classes.
"""
# The name of the method attribute that specifies it is abstract
ABSTRACT_METHOD_ATTRIBUTE: str = "__isabstractmethod__"

# The name of the class attribute that defines the abstract methods
# for a class
ABSTRACT_CLASS_ATTRIBUTE: str = "__abstractmethods__"
